/*
MIDI Processing for a Continuum 
 
*/

autowatch = 1;
inlets = 1;
outlets = 1;

var min_velocity = 10;
var max_velocity = 127;
var curve = 5;

var MidiClass = {
    UNKNOWN : 0,
    SYSTEM_REALTIME : 1,
    SYSTEM_COMMON : 2,
    VOICE : 3
};

function DataMapper(_min, _max, curve) {
	this.in_min = this.out_min = _min || 0;
	this.in_max = this.out_max = _max || 127;
	this.curve = curve || 1;
}

DataMapper.prototype.SetInMin = function(in_min) {
	this.in_min = in_min;
}

DataMapper.prototype.SetInMax = function(in_max) {
	this.in_max = in_max;
}

DataMapper.prototype.SetOutMin = function(out_min) {
	this.out_min = out_min;
}

DataMapper.prototype.SetOutMax = function(out_max) {
	this.out_max = out_max;
}

DataMapper.prototype.SetCurve = function(curve) {
	this.curve = curve;
}

DataMapper.prototype.Map = function(val) {
	return Math.round( ( Math.pow( ( ( val - this.in_min ) / (this.in_max - this.in_min) ), this.curve ) *
		(this.out_max - this.out_min) ) + this.out_min );
}

function MidiMsg( msg ) {
    this.msg = msg;
}

MidiMsg.prototype.Class = function() {
    if( this.msg[0] >= 0xF8 ) {
        return MidiClass.SYSTEM_REALTIME;
    } else if( this.msg[0] >= 0xF0 ) {
        return MidiClass.SYSTEM_COMMON;
    } else if( this.msg[0] >= 0x80 ) {
        return MidiClass.VOICE;
    } else {
        return MidiClass.UNKNOWN;
    }
}

MidiMsg.prototype.Channel = function() {
    if( this.Class() == MidiClass.VOICE ) {
        return this.msg[0] & 0xf;
    }
    return -1;
}

MidiMsg.prototype.SetChannel = function(chan) {
    if( this.Class() == MidiClass.VOICE ) {
        this.msg[0] = (this.msg[0] & 0xf0) | (chan & 0xf);
		return true;
    }
    return false;
}

MidiMsg.CreateController = function(ctrl, value, chan) {
	chan = chan || 0
	return new MidiMsg( [ (0xB0 | (chan & 0xf)), (ctrl & 0x7f), (value & 0x7f) ] );
}

MidiMsg.prototype.IsController = function() {
	return (this.msg[0] & 0xF0) == 0xB0;
}

MidiMsg.prototype.Controller = function() {
	if( this.IsController() ) {
		return this.msg[1];
	}
	return -1;
}

MidiMsg.prototype.ControllerValue = function() {
	if( (this.msg[0] & 0xF0) == 0xB0 ) {
		return this.msg[2];
	}
	return -1;
}

MidiMsg.CreatePitchwheel = function(value, chan) {
	chan = chan || 0
	value += 8192;
	return new MidiMsg( [ 0xE0 | (chan & 0xf), value & 0x7f, (value >> 7) & 0x7f ] );
}

MidiMsg.prototype.IsPitchWheel = function() {
	return this.msg[0] & 0xF0 == 0xE0;
}

MidiMsg.prototype.PitchWheel = function() {
	if( this.IsPitchWheel() ) {
		return ( this.msg[1] | ( this.msg[2] << 7 ) ) - 8192;
	}
	return -1;
}

MidiMsg.CreateNoteOn = function(key, vel, chan) {
	chan = chan || 0
	return new MidiMsg( [ 0x90 | (chan & 0xf), key & 0x7f, vel & 0x7f ] );
}

MidiMsg.CreateNoteOff = function(key, vel, chan) {
	chan = chan || 0
	vel = vel || 127
	return new MidiMsg( [ 0x80 | (chan & 0xf), key & 0x7f, vel & 0x7f ] );
}

MidiMsg.prototype.IsNote = function() {
	return (this.msg[0] & 0xE0) == 0x80;
}

MidiMsg.prototype.SetNoteVel = function(vel) {
	if( this.IsNote() ) {
		this.msg[2] = vel & 0x7f;
	}
}

MidiMsg.prototype.IsNoteOn = function() {
	return ((this.msg[0] & 0xF0) == 0x90) && (this.msg[2] != 0);
}

MidiMsg.prototype.IsNoteOff = function() {
	return ( (this.msg[0] & 0xF0) == 0x80 ) ||
		(((this.msg[0] & 0xF0) == 0x90) && (this.msg[2] == 0));
}

MidiMsg.prototype.NoteOnKeyVel = function() {
	if( ( (this.msg[0] & 0xF0) == 0x90 ) && (this.msg[2] != 0) ) {
		return this.msg.slice(1,3);
	}
	return [-1, -1];
}

MidiMsg.prototype.NoteOffKeyVel = function() {
	if( (this.msg[0] & 0xF0) == 0x80 ) {
		return this.msg.slice(1,3);
	} else if( ( (this.msg[0] & 0xF0) == 0x90 ) && (this.msg[2] == 0) ) {
		return [ this.msg[1], 127 ];
	}
	return [-1, -1];
}

function MidiAccumulator() {
    this._offset = 0;
    this._length = 0;
    this._msg = [];
}

MidiAccumulator.prototype._MidiSysCommonBytes = [ -1, 2, 3, 2, -1, -1, 1, -2 ];

MidiAccumulator.prototype.HandleMessage = function( msg ) {
    post( "This should be replaced with user code" );
	post();
}

MidiAccumulator.prototype.HandleError = function( msg ) { }

// Handle the high level control logic for accumulating a MIDI message
MidiAccumulator.prototype.Accumulate = function( val ) {

    // Realtime Messages
    if( val >= 0xF8 ) {
        this._NewStatus( val, 0 );

    // System Common Messages
    } else if( val >= 0xF0 ) {
        this._NewStatus( val, this._MidiSysCommonBytes[ v & 0xf ] );

    // 2-byte Voice Commands
    } else if( val >= 0xC0 && val <= 0xDF ) {
        this._NewStatus( val, 2 );

    // 3-byte Voice Commands
    } else if( val >= 0x80 ) {
        this._NewStatus( val, 3 );

    // Any other message just needs to be accumulated
    } else {
        this._AccumMultiByte( val );
    }
}

// Accumulate a byte in a multi-byte message
MidiAccumulator.prototype._AccumMultiByte = function( val ) {

    // Length -1 = In SysEx and accumulate forever 
    if( this._length == -1 ) {
        this._msg.push( val );

    // Length 2+ = Accumulate another message with the current status
    } else if( this._length > 1 ) {

        // Inject the byte into the multi-byte buffer
        this._msg[this._offset] = val;
        this._offset++;

        // If we hit the correct numer of bytes, handle the message
        if( this._offset == this._length ) {
            this.HandleMessage( new MidiMsg( this._msg ) );
            this._offset = 1;
        }

    // Any other length is an error
    } else {
        this.HandleError( [val] );
    }
}

// Handle the boundaries relating to a new status message
MidiAccumulator.prototype._NewStatus = function ( val, len ) {

    // Length 0 = System Realtime (doesn't effect the current status)
    if( len == 0 ) {
        this.HandleMessage( new MidiMsg( [val] ) );

    // Length -2 = End of a SysEx
    } else if( len == -2 ) {
        this._AccumMultiByte( val );
        this.HandleMessage( this._msg );
        this._msg.length = 1;

    // Handle real status changes here
    } else {
        this._msg = [val];
        this._offset = 1;
        this._length = len;

        // Length 1 = Handle immediately (no data required)
        if( len == 1 ) {
            this.HandleMessage( new MidiMsg( this._msg ) );
            this._length = 1;

        // Any other length means we need to accumulate messages
        } else if( len > 1 ) {
            this._msg.length = len;
        }
    }
}

var accum = new MidiAccumulator();
accum.mode = 0;
accum.cur_channel = -1;
accum.chan_data = Array.apply(null, new Array(15)).map(function(x){return [];});
accum.keys_on = [];
accum.iter = 0
accum.vel_mapper = new DataMapper(1,127,1);
accum.vel_mapper.SetOutMin(10);
accum.vel_mapper.SetCurve(0.5);

accum.SetMode = function( str_mode ) {
	if( str_mode == "Performance" ) {
		this.mode = 0;
	} else if( str_mode == "Patcher" ) {
		this.mode = 1;
	}
}

accum.SendMappedMessage = function( prefix, msg ) {

		// Force all channeld messages to channel 0, but on the proper instance
		if( msg.Channel() != -1 ) {
		
			// Remap the velocity
			if( msg.IsNoteOn() ) {
				msg.SetNoteVel( this.vel_mapper.Map( msg.NoteOnKeyVel()[1] ) );
			}
	
			// Force everything to channel 0
			msg.SetChannel(0);
		}
		
		// Send out the message
		outlet(0, prefix.concat( msg.msg ) );
}

accum.HandleMessage = function( msg ) {
	if( msg.Channel() == 15 ) { return; }
	//post(this.iter++); post("HandleMessage: "); post(msg.msg); post(" - "); post(this.mode); post();
	
	orig_chan = msg.Channel()
	
	// Performance mode, we pass channel to instance
	if( this.mode == 0 ) {
		
		// If the middi channel isn't important, broadcast to all instances
		if( orig_chan == -1 ) {
			this.SendMappedMessage( ["midievent"], msg );
		} else {	
			this.SendMappedMessage( ["instance", msg.Channel(), "midievent"], msg );
		}

	// Patch mode, we need to fake stuff all to channel 1
	} else if( this.mode == 1 ) {
		
		// Non-channel related messages
		if( msg.Class() != MidiClass.VOICE ) {
			this.SendMappedMessage( ["instance", 0, "midievent"], msg );
		
		// If we are the current channel, handle any events
		} else if( msg.Channel() == this.cur_channel ) {
			
			// See if we got a key on, and add it
			if( (key = msg.NoteOnKeyVel()[0]) != -1 ) {
				if( this.keys_on.indexOf(key) == -1 ) {
					this.keys_on.push( key );
				}
			}

			// See if we got a key off, and delete it
			if( (key = msg.NoteOffKeyVel()[0]) != -1 ) {
				if( (index = this.keys_on.indexOf(key)) != -1 ) {
					this.keys_on.splice(index, 1);
				}
			}

			// Send the event
			this.SendMappedMessage( ["instance", 0, "midievent"], msg );
			
		// If we have a note on on another channel, issue note offs for all
		// notes on the current channel
		} else if( msg.IsNoteOn() ) {
			
			// Destination channel
			new_chan = msg.Channel();
			
			// Do note offs
			this.keys_on.forEach( function(key) {
				this.SendMappedMessage( ["instance", 0, "midievent"], MidiMsg.CreateNoteOff(key) );
			}, this );
			this.keys_on = [ msg.NoteOnKeyVel()[0] ];
			
			// Do any accumulated channel messages
			this.chan_data[new_chan].forEach( function(acc_msg) {
				this.SendMappedMessage( ["instance", 0, "midievent"], acc_msg);
			}, this );
			this.chan_data[new_chan] = [];
			
			// Set the current channel
			this.cur_channel = new_chan;
			
			// Call back now that we're the current channel
			this.HandleMessage( msg );
			
		// If we have a note off, we can throw away any data we have been
		// keeping.
		} else if( msg.IsNoteOff() ) {
			this.chan_data[msg.Channel()] = []
		
		// Anything not on channel 16 should be accumulated to the channel
		// data.
		} else if( msg.Channel() != 15 ) {
			this.chan_data[msg.Channel()].push( msg );
		}			
	} else {
		post("Error: invalid mode.");
		post();
	}
}

accum.HandleError = function( msg ) {
	post( "Error" );
	post( msg );
	post();
}

function anything() {
	val = arrayfromargs(messagename,arguments);
	post("anything was called: " + val.toSource());
	post();
}

function mode(m) {
	accum.SetMode(m);
}

function msg_int(v) {
    accum.Accumulate(v);
}
