autowatch = 1;
inlets = 3;
outlets = 3;

var TEXT_INLET = 0;
var FONT_INLET = 1;
var WIDTH_INLET = 2;

var LINE_OUTLET = 0;
var DUMP_OUTLET = 1;
var SCROLL_OUTLET = 2;

setinletassist(0, "append/list/text: append text; set: clear, then append text");
setinletassist(1, "font, fontname, fontsize, lineheight");
setinletassist(2, "wrap width in pixels");

setoutletassist(0, "commands to 1st inlet of jit.cellblock (clearing and setting text)");
setoutletassist(1, "commands to 2nd inlet of jit.cellblock (scrolling)");

// scrolling doesn't happen if you do it right away, but any amount of delay seems to work
var SCROLL_DELAY = 1;
var scrollTask = new Task(scrollTaskImpl, this);

var debug = false;

var fontName = "Consolas";
var fontSize = 12;
var boxWidth = 550;
var lineNo = 0;
var lastLine = "";

var bodyText = "";

// initialize sketch object
sketch.default2d();
sketch.font(fontName);
sketch.fontsize(fontSize);

// sets/appends text
var set = list;
var text = list;
var append = list;
function list()
{
  if (inlet != TEXT_INLET) return;
  if ("set" == messagename) clear();
  var args = Array.prototype.slice.call(arguments);
  var newText = args.join(" ");
  bodyText += newText;
  setText(lastLine + newText);
  scrollTo(lineNo);  
}

setText.local = 1;
function setText(text) {
  var lines = wrapText(text, boxWidth);
  outputLines(lines);
}

// sets wrap width
function msg_int(width)
{
  if (inlet != WIDTH_INLET) return;
  boxWidth = width;
  reset();
  setText(bodyText);
}

// sets font
function font() {
  if (arguments.length > 0) {
    fontName = arguments[0];
    if (arguments.length > 1) {
      fontSize = arguments[1];
    }
    setFont();
  }
}

// sets font name
function fontname() {
  if (arguments.length > 0) {
    fontName = arguments[0];
    setFont();
  }
}

// sets font size
function fontsize() {
  if (arguments.length > 0) {
    fontSize = arguments[0];
    setFont();
  }  
}

function lineheight() {
  outlet(LINE_OUTLET, ["rowheight", arguments[0]]);
}

// called when font is changed
setFont.local = 1;
function setFont() {
  sketch.font(fontName);
  sketch.fontsize(fontSize);
  reset();
  setText(bodyText);
}

// called every time text is sent
setFont.local = 1;
function sendFont() {
  outlet(LINE_OUTLET, ["sendbox", "fontname", fontName]);
  outlet(LINE_OUTLET, ["sendbox", "fontsize", fontSize]);
}

function dump() {
  outlet(DUMP_OUTLET, ["list", bodyText]);
}

// clears text
function clear() {
  reset();
  bodyText = "";
}

// same as clear, but doesn't lose the text
// used when re-wrapping text after width or font changes
reset.local = 1;
function reset() {
  outlet(LINE_OUTLET, ["clear","all"]);
  outlet(LINE_OUTLET, ["rows", 0]);
  lineNo = 0;
  lastLine = "";  
}

function outputLines(lines)
{
  var numLines = lines.length;
  if (0 == numLines) return;
  sendFont();
  outlet(LINE_OUTLET, ["rows", lineNo + numLines]);
  for (var i = 0; i < numLines; i++) {
    outlet(LINE_OUTLET, ["set", 0, lineNo + i, lines[i]]);
  }
  lineNo += numLines - 1;
  lastLine = lines[numLines - 1];
}

function scrollTo(lineNo)
{
  scrollTask.arguments = [lineNo];
  scrollTask.schedule(SCROLL_DELAY);
}

function scrollTaskImpl()
{
  outlet(SCROLL_OUTLET, ["sync", "click", 1, arguments[0], 0, 0]);
}

function wrapText(text, width) {
  var paragraphs = text.split(/\n/);
  var numParagraphs = paragraphs.length;
  var wrapped = [];
  for (i = 0; i < numParagraphs; i++) {
    var p = paragraphs[i] + (i + 1 == numParagraphs ? "" : "\n");
    wrapped = wrapped.concat(wrapParagraph(p, width));
  }
  return wrapped;
}

function wrapParagraph(text, width) {
  var breakFound = false;
  var regex = / +/g;
  var index = 0;
  var lineWidth = 0;
  while (!breakFound) {
    // look ahead to the next possible break
    var match = regex.exec(text);
    var noMoreBreaksPossible = !match;
    var newIndex = noMoreBreaksPossible ? text.length : regex.lastIndex;

    // how long is the line if we break there?
    line = text.substring(0, newIndex);
    lineWidth = textWidth(line);
    if (debug) {
      post("line: "); post(line); post();
      post(lineWidth + " > " + width + " ? " + (lineWidth > width)); post();
    }

    if (lineWidth > width) {
      // break at previous index, unless it's 0
      if (0 == index) index = newIndex;
      breakFound = true;
      if (debug) { post("break found; index = " + index); post(); }
    } else if (noMoreBreaksPossible) {
      index = newIndex;
      breakFound = true;
      if (debug) { post("no more breaks possible; putting the rest on last line"); post(); }
    } else {
      index = newIndex; // continue
      if (debug) { post("no break found yet; index = " + index); post(); }
    }
  }
  if (debug) { post("ok, breaking at " + index); post(); }
  var parts = breakText(text, index);
  if (1 == parts.length) {
    return parts;
  } else {
    var lines = wrapParagraph(parts[1], width);
    lines.unshift(parts[0]);
    return lines;
  }
}

function breakText(text, index) {
  return 0 == index || index >= text.length ? [text] : [text.substring(0, index), text.substring(index)]
}

function textWidth(text) {
  return sketch.gettextinfo(text)[0];
}


// TESTING

test.suite = {
  testBreakText: function() {
    var actual;
    actual = breakText("1234567", 0);
    assertEquals(1, actual.length);
    assertEquals("1234567", actual[0]);
    actual = breakText("1234567", 1);
    assertEquals(2, actual.length);
    assertEquals("1", actual[0]);
    assertEquals("234567", actual[1]);
    actual = breakText("1234567", 7);
    assertEquals(1, actual.length);
    assertEquals("1234567", actual[0]);
    actual = breakText("1234567", 6);
    assertEquals(2, actual.length);
    assertEquals("123456", actual[0]);
    assertEquals("7", actual[1]);
  },
  testWrapText: function() {
    var actual;
    actual = wrapText("hello, world", 500);
    assertEquals("hello, world", actual[0], "500");
    actual = wrapText("hello, world", 5);
    assertEquals("hello, ", actual[0]);
    assertEquals("world", actual[1]);
    actual = wrapText("hello,\ndamn world", 5);
    assertEquals("hello,\n", actual[0]);
    assertEquals("damn ", actual[1]);
    assertEquals("world", actual[2]);
    //debug = true;
    actual = wrapText("I have of late, but wherefore I know not, lost all my mirth; foregone all custom of exercizes and indeed it goes so heavily with my disposition that this goodly frame, the earth, seems to me a sterile promontory.", 300);
    assertEquals(4, actual.length);
    assertEquals("I have of late, but wherefore I know not, lost all my mirth; foregone ", actual[0]);
    assertEquals("promontory.", actual[3]);
  }
};

function test() {
  for (f in test.suite) {
    post("running " + f); post();
    (test.suite[f])();
  }
  post("tests done"); post();
}

function assert(condition, message) {
  if (!message) message = "";
  if (!condition) {
    message = "Assertion Failed! " + message;
    post(message);
    post();
  }
  return;
}

function assertEquals(expected, actual, message) {
  message = message ? message + ": " : "";
  message += "expected |" + expected + "|, got |" + actual + "|";
  assert(expected == actual ? true : false, message);
}